!============================================================
! sdkLCS.clw -- List Column Sorter class implementation
!
!  2010 Simon Kemp, simon@sdk.demon.co.uk
!
!  This software is provided 'as-is', without any express or implied
!  warranty.  In no event will the author be held liable for any damages
!  arising from the use of this software.
!
!  Permission is granted to anyone to use this software for any purpose,
!  including commercial applications, and to alter it and redistribute it
!  freely, subject to the following restrictions:
!
!  1. The origin of this software must not be misrepresented; you must not
!     claim that you wrote the original software. If you use this software
!     in a product, an acknowledgment in the product documentation would be
!     appreciated but is not required.
!
!  2. Altered source versions must be plainly marked as such, and must not be
!     misrepresented as being the original software.
!
!  3. This notice may not be removed or altered from any source distribution.
!
!============================================================
    member()
    include('sdklcs.inc'),once
    include('keycodes.clw'),once

  compile('***',_C60_)
    map
      SDKLCSCompareFunc(*group, *group),signed
      module('clarionrtl')
        SDKLCSSortFreeRTL(*queue, *SDKLCSCompareFunc),NAME('Cla$FREEQueueFKey')
      end
    end
  ***
  omit('***',_C60_)
    map
      SDKCompareFuncType(*group, *group),signed,type
      SDKLCSCompareFunc(*group, *group),signed
      module('clarionrtl')
        SDKSortC55(*queue, SDKCompareFuncType),NAME('Cla$SORTqueuefkey')
      end
    end
  ***
    map 
      set(*queue q,long setat=0)
      next(*queue q),long,proc
      previous(*queue q),long,proc
      copy(*queue sq, *queue tq, long append=0)
      module('win32') 
        OutputDebugString(*cstring lpOutputString),pascal,raw,name('OutputDebugStringA')
      end 
    end 
    
sdkLCS.Init  procedure(long thelistfeq, *queue listqueue, long tweaklineheight=0)
    code
    self.initcalled_ = true 
    free(self.columns)
    free(self.keys)
    self.alertkeystd = MouseLeft
    self.alertkeymulti = CtrlMouseLeft
    self.alertkeyclear = ShiftMouseLeft
    self.headereffects = false
    self.multikey = true
    self.prependindicators = false
    self.showcursoronsort = true
    self.sortonchanges = true
    self.listfeq = thelistfeq
    self.listfrom &= listqueue
    if tweaklineheight <> 0
      self.listfeq{PROP:LineHeight} = self.listfeq{PROP:LineHeight} + tweaklineheight
    end
    self.listfeq{PROP:Alrt, 255} = self.alertkeystd
    self.listfeq{PROP:Alrt, 255} = self.alertkeymulti
    self.listfeq{PROP:Alrt, 255} = self.alertkeyclear
    self.stickysort &= null
    clear(self.columns)
    loop
      self.columns.column += 1
      if not self.listfeq{PROPLIST:Exists, self.columns.column}
        break
      end
      self.columns.headertext = self.listfeq{PROPLIST:Header, self.columns.column}
      self.columns.fieldno = self.listfeq{PROPLIST:FieldNo, self.columns.column}
      if self.listfeq{PROP:VLBval} = false
        self.columns.columnwho = who(self.listfrom, self.columns.fieldno)
        self.columns.isastring = isstring(what(self.listfrom, self.columns.fieldno))
      end 
      add(self.columns)
    end
    if records(self.columns)
      self.listfeq{PROPLIST:Resize,records(self.columns)} = false
    end
    self.blockcolumnoptions = false
    if self.regionfeq = 0
      self.regionfeq = create(0,CREATE:region)
      self.regionfeq{prop:imm} = true
    end

sdkLCS.InitCalled  procedure!,long 
    code 
    return self.initcalled_

sdkLCS.ChangeQueue  procedure(*queue listqueue)
    code 
    self.listfrom &= listqueue
    !self.TakeEvent

sdkLCS.Kill  procedure
    code
    free(self.keys)
    self.ShowHeaders
    free(self.columns)
    self.listfrom &= null

sdkLCS.SetColumnOptions  procedure(<string nosorting>, <string casesensitive>, <string restrictsort>, <string overridecolumnfields>)
myPtr1  long
myAny   any
    code
    if self.blockcolumnoptions = true
      return
    end
    self.blockcolumnoptions = true
    loop myPtr1 = 1 to len(clip(overridecolumnfields)) - 5 by 6
      self.columns.column = overridecolumnfields[myPtr1 : myPtr1 + 2]
      get(self.columns, +self.columns.column)
      if not errorcode()
        myAny &= null
        myAny &= what(self.listfrom,overridecolumnfields[myPtr1 + 3 : myPtr1 + 5])
        if not myAny &= null
          self.columns.fieldno = overridecolumnfields[myPtr1 + 3 : myPtr1 + 5]
          self.columns.columnwho = who(self.listfrom, self.columns.fieldno)
          self.columns.isastring = isstring(what(self.listfrom, self.columns.fieldno))
          put(self.columns)
        end
      end
    end
    loop self.columns.column = 1 to len(clip(nosorting))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if nosorting[self.columns.column] = true
        self.columns.sortoptions += 2^0 ! no sorting
        put(self.columns)
      end
    end
    loop self.columns.column = 1 to len(clip(casesensitive))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if casesensitive[self.columns.column] = true
        self.columns.sortoptions += 2^1 ! case-sensitive
        put(self.columns)
      end
    end
    loop self.columns.column = 1 to len(clip(restrictsort))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if inrange(restrictsort[self.columns.column],1,3)
        execute restrictsort[self.columns.column]
          self.columns.sortoptions += 2^2 ! restrict +
          self.columns.sortoptions += 2^3 ! restrict -
          begin
            if not band(self.columns.sortoptions, 2^0)
              self.columns.sortoptions += 2^0 ! no sorting
            end
          end
        end
        put(self.columns)
      end
    end

sdkLCS.SetStrings  procedure(string poverride)    
    code    
    loop self.columns.column = 1 to len(clip(poverride))
      get(self.columns, +self.columns.column)
      if errorcode()
        cycle
      end
      if poverride[self.columns.column] = true
        self.columns.isastring = true
      else
        self.columns.isastring = false
      end
      put(self.columns)
    end

sdkLCS.SetInitialSort  procedure(<*string variablesort>, <string initialsort>, long locked=0)
    code
    if not omitted(2)
      self.stickysort &= variablesort
      self.initialsort = left(variablesort)
    else
      self.initialsort = left(initialsort)
    end
    self.locked = locked

sdkLCS.SetLocator  procedure(long pFEQ, *string pUse, long pPromptFEQ=0, long pIncremental=0)
myPtr   long 
    code 
    self.locatorfeq = pFEQ
    self.locatoruse &= pUse 
    self.locatorfeq{PROP:Upr} = 1
    self.locatorpromptfeq = pPromptFEQ
    self.locatorincremental = pIncremental
    if self.alertedlocator = false
      self.alertedlocator = true 
      loop myPtr = 1 to size(sdkLCSLocate) 
        self.listfeq{PROP:Alrt, 255} = val(sdkLCSLocate[myPtr]) 
      end  
    end   

sdkLCS.SetTagColumn  procedure(long pNumber, *string pData, *long pIcon, long pClearFEQ=0)  
    code 
    self.tagcolumn = pNumber
    self.tagdata &= pData
    self.tagicon &= pIcon
    self.tagclearfeq = pClearFEQ
    
sdkLCS.Start procedure
myPtr1          long
myPtr2          long
    code
    self.blockcolumnoptions = true
    if self.initialsort
      case self.initialsort[1]
      of '+' orof '-'
      else
        self.initialsort = '+' & self.initialsort
      end
      myPtr1 = 1
      loop while myPtr1 < len(clip(self.initialsort))
        loop myPtr2 = myPtr1 + 1 to len(clip(self.initialsort)) + 1
          clear(self.keys)
          if not inrange(val(self.initialsort[myPtr2]),val('0'),val('9'))
            self.keys.level = records(self.keys) + 1
            self.keys.column = deformat(self.initialsort[myPtr1 : myPtr2 - 1])
            if self.keys.column < 0
              self.keys.direction = -1
            else
              self.keys.direction = +1
            end
            self.keys.column = abs(self.keys.column)
            self.columns.column = self.keys.column
            get(self.columns,+self.columns.column)
            if not errorcode() and not band(self.columns.sortoptions,2^0) and not (band(self.columns.sortoptions,2^2) and self.keys.direction < 0) and not (band(self.columns.sortoptions,2^3) and self.keys.direction > 0)
              self.keys.fieldno = self.columns.fieldno
              self.keys.sortoptions = self.columns.sortoptions
              self.keys.isastring = self.columns.isastring
              add(self.keys)
            end
            break
          end
        end
        myPtr1 = myPtr2
      end
      if not self.multikey
        loop while records(self.keys) > 1
          get(self.keys,records(self.keys))
          delete(self.keys)
        end
      end
    end
    if not self.stickysort &= null
      self.stickysort = self.showsort()
    end
    self.ShowHeaders
    self.SortQ

sdkLCS.SetSort  procedure(long ps1, long ps2=0, long ps3=0)!,long,proc
myptr   long
mysort  long
    code
    free(self.keys)
    loop myptr = 1 to 3
      execute myptr
        mysort = ps1
        mysort = ps2
        mysort = ps3
      end
      clear(self.keys)
      self.keys.level = records(self.keys) + 1
      if mysort < 0
        self.keys.direction = -1
      else
        self.keys.direction = +1
      end
      self.keys.column = abs(mysort)
      self.columns.column = self.keys.column
      get(self.columns,+self.columns.column)
      if not errorcode() and not band(self.columns.sortoptions,2^0) and not (band(self.columns.sortoptions,2^2) and self.keys.direction < 0) and not (band(self.columns.sortoptions,2^3) and self.keys.direction > 0)
        self.keys.fieldno = self.columns.fieldno
        self.keys.sortoptions = self.columns.sortoptions
        self.keys.isastring = self.columns.isastring
        add(self.keys)
      end
    end
    self.ShowHeaders
    return true

sdkLCS.NewHeader  procedure(long pcolumnnumber)  
    code 
    clear(self.columns) 
    self.columns.column = pcolumnnumber
    get(self.columns,+self.columns.column)
    if not errorcode()
      self.columns.headertext = self.listfeq{PROPLIST:Header, self.columns.column}
      put(self.columns)  
    end
    self.ShowHeaders

sdkLCS.NewHeader  procedure(long pcolumnnumber, string pheadertext)  
    code 
    clear(self.columns) 
    self.columns.column = pcolumnnumber
    get(self.columns,+self.columns.column)
    if not errorcode()
      self.columns.headertext = pheadertext
      put(self.columns)  
    end
    self.ShowHeaders

sdkLCS.ShowHeaders  procedure(long reset=0, long reload=0)
myPtr   long
    code
    loop myPtr = 1 to records(self.columns)
      get(self.columns, myPtr)
      if reset
        self.listfeq{PROPLIST:Header, self.columns.column} = self.columns.headertext
      else
        if reload
          self.columns.headertext = self.listfeq{PROPLIST:Header, self.columns.column}
          put(self.columns)
        end
        self.keys.column = self.columns.column
        get(self.keys, +self.keys.column)
        if errorcode()
          self.listfeq{PROPLIST:Header, self.columns.column} = self.columns.headertext
        elsif self.prependindicators
          self.listfeq{PROPLIST:Header, self.columns.column} = '[' & choose(self.keys.direction < 0,'-','+') & choose(records(self.keys)=1,'',clip(self.keys.level)) & ']' & left(self.columns.headertext)
        else
          self.listfeq{PROPLIST:Header, self.columns.column} = clip(self.columns.headertext) & '[' & choose(self.keys.direction < 0,'-','+') & choose(records(self.keys)=1,'',clip(self.keys.level)) & ']'
        end
      end
    end

sdkLCS.ShowSort  procedure ! ,string
myPtr       long
myResult    string(1024)
    code
    sort(self.keys, +self.keys.level)
    loop myPtr = 1 to records(self.keys)
      get(self.keys, myPtr)
      if self.keys.direction < 0
        myResult = clip(myResult) & '-' & self.keys.column
      else
        myResult = clip(myResult) & '+' & self.keys.column
      end
    end
    return clip(myResult)

sdkLCS.SetNoSorting  procedure(long pcolumnnumber, long pnosorting=1)!,long,proc
    code
    self.columns.column = pcolumnnumber
    get(self.columns, +self.columns.column)
    if not errorcode()
      if pnosorting
        self.columns.sortoptions = bor(self.columns.sortoptions, 2^0) ! no sorting
      elsif band(self.columns.sortoptions, 2^0)
        self.columns.sortoptions -= 2^0
      end
      put(self.columns)
      return true
    end
    return false

sdkLCS.ColumnIsSorted  procedure(long pcolumnnumber)!,long
    code
    self.keys.column = pcolumnnumber
    get(self.keys, +self.keys.column)
    if errorcode()
      return false
    else
      return true
    end

sdkLCS.SortQ procedure(long dontrelocate=0)
mySaveQ &string
myPtr   long
myAny1  any
myAny2  any
    code
    if self.listfrom &= null
      return
    end
    if records(self.listfrom) and records(self.keys)
      if self.showcursoronsort
        setcursor(CURSOR:Wait)
      end
      if dontrelocate = 0
        mySaveQ &= new string(size(self.listfrom))
        myPtr = choice(self.listfeq)
        if myPtr
          get(self.listfrom, myPtr)
          mySaveQ = self.listfrom
        end
      end
      sort(self.keys, +self.keys.level)
      SDKLCSkeyqref &= self.keys
  compile('***',_C60_)
      SDKLCSSortFreeRTL(self.listfrom, SDKLCSCompareFunc)
      sort(self.listfrom, SDKLCSCompareFunc)
  ***
  omit('***',_C60_)
      get(self.listfrom, 1)
      add(self.listfrom, 1)
      delete(self.listfrom)
      SDKSortC55(self.listfrom, SDKLCSCompareFunc)
  ***
      if dontrelocate = 0
        if myPtr
          loop myPtr = 1 to records(self.listfrom)
            get(self.listfrom, myPtr)
            if mySaveQ = self.listfrom
              self.listfeq{PROP:Selected} = myPtr
              break
            end
          end
        end
        dispose(mySaveQ)
      end
      if self.columnoneislocator
        myAny1 &= what(self.listfrom,1)
        get(self.keys,1)
        myAny2 &= what(self.listfrom,self.keys.fieldno)
        loop myPtr = 1 to records(self.listfrom)
          get(self.listfrom,myPtr)
          myAny1 = myAny2
          put(self.listfrom)
        end
      end
      if self.showcursoronsort
        setcursor
      end
    elsif self.columnoneislocator
      myAny1 &= what(self.listfrom,1)
      loop myPtr = 1 to records(self.listfrom)
        get(self.listfrom,myPtr)
        myAny1 = ''
        put(self.listfrom)
      end
    end
    self.Storechanges = changes(self.listfrom)

sdkLCS.GetPosition  procedure(string pLocate)!,long 
myResult    long 
myColumn    any 
    code 
    if not self.listfrom &= null and records(self.listfrom) and records(self.keys)
      sort(self.keys, +self.keys.level)
      SDKLCSkeyqref &= self.keys
      clear(self.listfrom) 
      get(self.listfrom, 0) 
      get(self.keys, 1) 
      myColumn &= what(self.listfrom, self.keys.fieldno) 
      myColumn = pLocate 
      myResult = position(self.listfrom)
    end   
    return myResult

sdkLCS.PropSetter  procedure(long pFEQ, long pProp, string pValue) 
    code 
    case pProp 
    of PROP:Disable orof PROP:Hide
      if pValue <> '1'
        pValue = ''
      end
    end
    if pFEQ{pProp} <> pValue
      pFEQ{pProp} = pValue
    end   

sdkLCS.Debug  procedure(string pWhat) 
OutputString    &cstring
    code
    if pWhat
      outputstring &= new cstring(len(clip(pWhat))+1+9)
      outputstring = '[sdklcs] ' & pWhat 
      OutputDebugString(OutputString)
      dispose(OutputString)
    end

sdkLCS.TakeEvent  procedure ! ,long
myPtr       long 
myKeyPress  string(1) 
myLo        long 
myHi        long   
    code
    if self.listfrom &= null
      case field()
      of self.listfeq
        case event()
        of EVENT:PreAlertKey
          return true
        end
      end
      return false
    end
    if self.sortonchanges
      if self.storechanges <> changes(self.listfrom)
        self.SortQ
        return false
      end
    end
    case field()
    of self.tagclearfeq
      case event() 
      of EVENT:Accepted
        self.ClearTags
      end 
    of self.locatorfeq 
      case event() 
      of EVENT:Selected
        self.locatorfeq{PROP:Touched} = true
      of EVENT:Accepted
        if 0{PROP:AcceptAll} = false 
          myPtr = self.GetPosition(self.locatoruse) 
          if myPtr 
            if myPtr > records(self.listfrom) 
              myPtr = records(self.listfrom) 
            end 
            select(self.listfeq, myPtr) 
            post(EVENT:NewSelection, self.listfeq) 
          end 
        end 
      end   
    of self.listfeq
      case event()
      of EVENT:NewSelection
        if self.listfeq{PROP:Column} > 0 
          if self.listfeq{PROPLIST:Width, self.listfeq{PROP:Column}} > 0
            self.lastcolumn = self.listfeq{PROP:Column}
          elsif self.lastcolumn and self.listfeq{PROPLIST:Width, self.lastcolumn} > 0
            self.listfeq{PROP:Column} = self.lastcolumn
          end
        end
        if self.tagcolumn and self.listfeq{PROPLIST:MouseUpField} = self.tagcolumn and self.listfeq{PROPLIST:MouseUpZone} = LISTZONE:Icon
          case keycode()
          of MouseLeft ! orof CtrlMouseLeft <- just confusing?
            setkeycode(0)
            get(self.listfrom, choice(self.listfeq))
            case self.tagicon
            of sdkLCS::Tagged
              self.tagdata = ''
              self.tagicon = sdkLCS::NotTagged
              self.lasttagpointer = pointer(self.listfrom) 
              self.lasttagaction = false
            of sdkLCS::NotTagged
              self.tagdata = true 
              self.tagicon = sdkLCS::Tagged
              self.lasttagpointer = pointer(self.listfrom) 
              self.lasttagaction = true
            end 
            put(self.listfrom)
          of ShiftMouseLeft  
            setkeycode(0)
            if self.lasttagpointer
              myLo = self.lasttagpointer
              myHi = choice(self.listfeq) 
              self.SortTwoThings(myLo, myHi) 
              loop myPtr = myLo to myHi 
                get(self.listfrom, myPtr) 
                case self.tagicon
                of sdkLCS::BlockTagging
                else 
                  case self.lasttagaction
                  of true 
                    self.tagdata = true 
                    self.tagicon = sdkLCS::Tagged
                  else 
                    self.tagdata = ''
                    self.tagicon = sdkLCS::NotTagged
                  end 
                  put(self.listfrom) 
                end   
              end 
              self.lasttagpointer = choice(self.listfeq)
            end   
          end   
        end 
      of EVENT:AlertKey
        if instring(chr(keycode()), sdkLCSLocate, 1, 1)
          if self.locatorfeq and self.locatorincremental and records(self.keys) and sub(upper(self.locatorfeq{PROP:Text}),1,2) = '@S'
            loop 1 times 
              case keycode() 
              of BSKey
                if len(self.incrementallocator) > 0 
                  self.incrementallocator = sub(self.incrementallocator, 1, len(self.incrementallocator)-1) 
                else 
                  break 
                end 
              else 
                if len(self.incrementallocator) = size(self.incrementallocator) - 1
                  break 
                end 
                case keycode() 
                of 189
                 myKeyPress = '-' 
                of 190
                  myKeyPress = '.' 
                of 191   
                  myKeyPress = '/' 
                of 192
                  myKeyPress = '''' 
                else   
                  myKeyPress = upper(chr(keycode())) 
                end                 
                self.incrementallocator = self.incrementallocator & myKeyPress
              end   
              change(self.locatorfeq, self.incrementallocator) !!
              myPtr = self.GetPosition(self.incrementallocator) 
              if myPtr 
                if myPtr > records(self.listfrom) 
                  myPtr = records(self.listfrom) 
                end   
                if myPtr <> choice(self.listfeq) 
                  select(self.listfeq, myPtr) 
                  post(EVENT:NewSelection, self.listfeq) 
                end 
              end  
            end   
          end   
        end 
      of EVENT:PreAlertKey
        case keycode()
        of self.alertkeystd orof self.alertkeymulti orof self.alertkeyclear
          if not self.locked
            if self.listfeq{PROPLIST:MouseDownRow} = 0 and self.listfeq{PROPLIST:MouseDownZone} = LISTZONE:header
              self.lastfield = self.listfeq{PROPLIST:MouseDownField}
              self.lastkeycode = keycode()
              if keycode() = self.alertkeyclear
                if self.lastfield
                  if self.headereffects = false or self.listfeq{PROP:ListFEQ} > 0 ! couldn't get prop:edit working with drop lists and combos
                    self.actionkeycode(keycode())
                  else
                    self.regiondown
                  end
                end
              else
                self.columns.column = self.lastfield
                get(self.columns, +self.columns.column)
                if not errorcode() and not band(self.columns.sortoptions,2^0)
                  if self.headereffects = false or self.listfeq{PROP:ListFEQ} > 0 ! couldn't get prop:edit working with drop lists and combos
                    self.actionkeycode(keycode())
                  else
                    self.regiondown
                  end
                else
                  clear(self.lastfield)
                end
              end
            end
          end
          return true
        end
      end
    of self.regionfeq
      if self.regionfeq{PROP:Hide}= false
        case event()
        of EVENT:MouseOut
          self.listfeq{PROP:Edit,self.lastfield} = 0
          self.regionfeq{PROP:Hide} = true
        of EVENT:MouseUp
          if self.regionfeq{PROP:Hide} = false
            self.listfeq{PROP:Edit,self.lastfield} = 0
            self.regionfeq{PROP:Hide} = true
            self.actionkeycode(self.lastkeycode)
          end
        end
      end
    end
    if self.locatorfeq
      if records(self.keys) = 0
        self.PropSetter(self.locatorfeq, PROP:Disable, true) 
        if self.locatorpromptfeq
          self.PropSetter(self.locatorpromptfeq, PROP:Disable, true) 
        end 
      else 
        get(self.keys, 1) 
        case sub(upper(self.listfeq{PROPLIST:Picture, self.keys.column}), 1, 2) 
        of '@D' 
          self.locatorfeq{PROP:Text} = '@D05B'
          do entrylocator 
        of '@T'
          self.locatorfeq{PROP:Text} = '@T01B'
          do entrylocator 
        else 
          self.locatorfeq{PROP:Text} = '@S' & len(self.locatoruse)
          if self.locatorincremental=true
            if self.locatoruse <> sub(self.incrementallocator, 1, size(self.locatoruse))
              change(self.locatorfeq, self.incrementallocator) 
            end   
            self.PropSetter(self.locatorfeq, PROP:Trn, true)
            self.PropSetter(self.locatorfeq, PROP:Disable, true) 
            if self.locatorpromptfeq
              self.PropSetter(self.locatorpromptfeq, PROP:Disable, true) 
            end 
          else 
            do entrylocator 
          end   
        end 
      end   
    end 
    if self.tagcolumn and self.tagclearfeq 
      self.PropSetter(self.tagclearfeq, PROP:Hide, choose(self.HasTags()=false))
    end 
    return false
    
entrylocator  routine 
    self.PropSetter(self.locatorfeq, PROP:Trn, false)
    self.PropSetter(self.locatorfeq, PROP:Disable, false) 
    if self.locatorpromptfeq
      self.PropSetter(self.locatorpromptfeq, PROP:Disable, false) 
    end 
    
sdkLCS.HasTags  procedure!,long 
myResult    long 
    code 
    if self.tagcolumn 
      self.tagdata = true 
      get(self.listfrom, '+' & who(self.listfrom, self.tagcolumn)) 
      if not errorcode() 
        myResult = true 
      end 
    end 
    return myResult 

sdkLCS.ClearTags  procedure 
    code 
    if self.tagcolumn 
      set(self.listfrom) 
      loop while next(self.listfrom) 
        if self.tagicon = sdkLCS::Tagged
          self.tagdata = ''
          self.tagicon = sdkLCS::NotTagged
          put(self.listfrom) 
        end 
      end 
      self.lasttagpointer = 0
    end   

sdkLCS.TagCount  procedure!,long 
myResult    long 
    code 
    if self.tagcolumn 
      set(self.listfrom) 
      loop while next(self.listfrom) 
        if self.tagicon = sdkLCS::Tagged
          myResult += 1
        end 
      end 
    end   
    return myResult 
    
sdkLCS.SortTwoThings  procedure(*? pThing1, *? pThing2)
any1    any
any2    any
    code
    any1 = pThing1
    any2 = pThing2
    if pThing1 > pThing2
      pThing1 = any2
      pThing2 = any1
    end
    
sdkLCS.UnSortQ procedure
    code
    if self.listfrom &= null
      return
    end
  compile('***',_C60_)
    SDKLCSSortFreeRTL(self.listfrom, SDKLCSCompareFunc)
  ***
  omit('***',_C60_)
    get(self.listfrom, 1)
    add(self.listfrom, 1)
    delete(self.listfrom)
  ***

sdkLCS.ActionKeyCode  procedure(long thekeycode) ! ,private
mysavesort  string(1024)
myptr       long
    code
    case thekeycode
    of self.alertkeystd orof self.alertkeymulti orof self.alertkeyclear
      mysavesort = self.ShowSort()
      if self.lastkeycode = self.alertkeyclear
        free(self.keys)
      else
        sort(self.keys, +self.keys.level)
        self.columns.column = self.lastfield
        get(self.columns, +self.columns.column)
        if not errorcode() and not band(self.columns.sortoptions,2^0)
          self.keys.column = self.columns.column
          get(self.keys, +self.keys.column)
          if errorcode()
            if not self.multikey or keycode() = self.alertkeystd
              free(self.keys)
            end
            clear(self.keys)
            self.keys.column = self.columns.column
            self.keys.direction = 1
            self.keys.fieldno = self.columns.fieldno
            self.keys.sortoptions = self.columns.sortoptions
            self.keys.isastring = self.columns.isastring
            add(self.keys)
          else
            self.keys.direction *= -1
            if not self.multikey or keycode() = self.alertkeystd
              free(self.keys)
              add(self.keys)
            else
              put(self.keys)
            end
          end
          if (band(self.keys.sortoptions,2^2) and self.keys.direction < 0) or (band(self.keys.sortoptions,2^3) and self.keys.direction > 0)
            self.keys.direction *= -1
            put(self.keys)
          end
        end
      end
      loop myPtr = 1 to records(self.keys)
        get(self.keys, myPtr)
        self.keys.level = myPtr
        put(self.keys)
      end
      if not self.stickysort &= null
        self.stickysort = self.showsort()
      end
      if mysavesort <> self.ShowSort()
        self.ShowHeaders
        self.SortQ
      end
    end

sdklcs.FieldInGroup  procedure ! ,long,private
lRow            long,auto
lGroup          long,auto
lIndex          long,auto
lLastColumns    long,auto
    code
    if self.listfeq{PROPLIST:GroupNo + PROPLIST:Group,self.lastfield} + 0 = 0
      return 0
    else
      lGroup = self.listfeq{PROPLIST:GroupNo,self.lastfield} + 0
      lRow   = 2
      lIndex = 0
      loop lIndex = self.lastfield to 1 by -1
        if lGroup <> self.listfeq{PROPLIST:GroupNo,lIndex} + 0
          lLastColumns = lIndex - 1
          break
        end
        if self.listfeq{PROPLIST:LastOnLine,lIndex}
           lRow+=1
        end
      end
      lIndex = self.lastfield + 1
      loop
        if lGroup <> self.listfeq{PROPLIST:GroupNo,lIndex} + 0
          lLastColumns=lIndex-1
          break
        end
        if self.listfeq{PROPLIST:LastOnLine,lIndex}
          lRow += 1
        end
        lIndex += 1
      end
      if self.listfeq{PROPLIST:LastOnLine,lLastColumns}
        lRow -= 1
      end
      if not self.listfeq{PROPLIST:Group + PROPLIST:Header,self.lastfield}
        lRow -= 1
      end
      return lRow
    end

sdkLCS.RowSelectedInGroup  procedure ! ,long,private
lRow        long,auto
lGroup      long,auto
lIndex      long,auto
lColumns    long,auto
    code
    if self.listfeq{PROPLIST:GroupNo + PROPLIST:Group,self.lastfield} + 0 = 0
      return 0
    else
      lGroup = self.listfeq{PROPLIST:GroupNo,self.lastfield} + 0
      lRow = 2
      lIndex = 0
      if self.listfeq{PROPLIST:LastOnLine,self.lastfield}
        lRow = 1
      end
      loop lIndex = self.lastfield to 1 by -1
        if lGroup <> self.listfeq{PROPLIST:GroupNo,lIndex} + 0
          break
        end
        if self.listfeq{PROPLIST:LastOnLine,lIndex}
          lRow += 1
        end
      end
      if not self.listfeq{PROPLIST:Group + PROPLIST:Header,self.lastfield}
        lRow -= 1
      end
      return lRow
    end

sdklcs.RegionDown  procedure
lPixels         long
lRow            long
lDelta          long
lScrollWidth    long
lOldControl     long
lx1             long
lx2             long
    code
    lPixels = 0{prop:pixels}
    0{prop:pixels} = true
    loop lx2 = self.listfeq{PROP:HScrollPos} + 1 to self.lastfield - 1
      lx1 += self.listfeq{PROPLIST:width,lx2}
    end
    self.regionfeq{PROP:XPos} = lx1
    self.regionfeq{PROP:Width} = self.listfeq{PROPLIST:width,self.lastfield}

    lOldControl = self.listfeq{PROP:Edit,self.lastfield}
    self.listfeq{PROP:Edit,self.lastfield} = self.regionfeq
    if lOldControl = self.regionfeq
      lOldControl = 0
    end
    self.listfeq{PROP:Edit,self.lastfield} = lOldControl
    lRow = self.RowSelectedInGroup()
    if lRow
      self.regionfeq{PROP:Height} = (self.listfeq{PROP:HeaderHeight} / self.FieldInGroup())
      self.regionfeq{PROP:YPos} = (lRow * self.regionfeq{PROP:Height})-self.regionfeq{PROP:Height}-1
      self.regionfeq{PROP:Height} = self.regionfeq{PROP:Height} + 1
    else
      self.regionfeq{PROP:Height} = self.listfeq{PROP:HeaderHeight} - 1
      self.regionfeq{PROP:YPos} = 0
    end
    if self.listfeq{PROPLIST:Icon,self.lastfield}
      lDelta = self.listfeq{PROPLIST:Width,self.lastfield} - self.regionfeq{PROP:Width}
      if lDelta<0
        lDelta = 15
      end
      self.regionfeq{PROP:XPos} = self.regionfeq{PROP:XPos} - lDelta
      self.regionfeq{PROP:Width} = self.regionfeq{PROP:Width} + lDelta
    end
    if self.listfeq{PROP:VScroll} and records(self.listfrom) > self.listfeq{PROP:Items}
      lScrollWidth = 20  ! getsystemmetrics perhaps......
    else
      lScrollWidth = 3
    end
    if (self.regionfeq{PROP:XPos}+self.regionfeq{PROP:Width}>=self.listfeq{PROP:Width}-lScrollWidth)
      self.regionfeq{PROP:Width}=self.regionfeq{PROP:Width}-(self.regionfeq{PROP:XPos}+self.regionfeq{PROP:Width}-(self.listfeq{PROP:Width}-lScrollWidth+1))
    end
    self.regionfeq{PROP:BevelInner} = -1
    self.regionfeq{PROP:BevelOuter} = -1
    self.regionfeq{PROP:YPos} = self.regionfeq{PROP:YPos}+self.listfeq{PROP:YPos}+1
    self.regionfeq{PROP:XPos} = self.regionfeq{PROP:XPos}+self.listfeq{PROP:XPos}+1
    self.regionfeq{PROP:Hide} = false
    0{prop:pixels} = lPixels

sdkLCS.Construct  procedure ! ,private
    code
    self.columns &= new sdkLCScolq
    self.keys &= new sdkLCSkeyq

sdkLCS.Destruct  procedure ! ,private
    code
    dispose(self.columns)
    dispose(self.keys)

SDKLCSCompareFunc procedure(*group cfTarget, *group cfCompare) ! ,signed
myPtr       long
myanyt      any
myanyc      any
    code
    loop myPtr = 1 to records(SDKLCSkeyqref)
      get(SDKLCSkeyqref, myPtr)
      myanyt &= what(cfTarget, SDKLCSkeyqref.fieldno)
      myanyc &= what(cfCompare, SDKLCSkeyqref.fieldno)
      if SDKLCSkeyqref.isastring and not band(SDKLCSkeyqref.sortoptions,2^1)
        ! case-less string sort required
        if upper(myanyt) < upper(myanyc)
          return -SDKLCSkeyqref.direction
        elsif upper(myanyt) > upper(myanyc)
          return SDKLCSkeyqref.direction
        end
      elsif myanyt < myanyc
        return -SDKLCSkeyqref.direction
      elsif myanyt > myanyc
        return SDKLCSkeyqref.direction
      end
    end
    return 0 ! all keys are equal


set  procedure(*queue q,long setat=0)
    code
    get(q,0)

next  procedure(*queue q)!,long,proc
    code
    get(q,pointer(q)+1)
    if errorcode()
      return false
    else
      return true
    end

previous  procedure(*queue q)!,long,proc
    code
    if pointer(q) = 0
      get(q,records(q))
    else
      get(q,pointer(q)-1)
    end
    if errorcode()
      return false
    else
      return true
    end

copy  procedure(*queue sq, *queue tq, long append=0)
    code
    if not append
      free(tq)
    end
    clear(tq)
    set(sq)
    loop while next(sq)
      tq = sq
      add(tq)
    end
